class tkanchor:
    from tkinter.constants import (W,E,EW,N,NW,NE,S,SE,SW,NSEW,CENTER,NS)
    '''
    W = tkcontants.W
    E = tkcontants.E
    EW = tkcontants.EW
    N = tkcontants.N
    NW = tkcontants.NW
    NE = tkcontants.NE
    S = tkcontants.S
    SE = tkcontants.SE
    SW = tkcontants.SW
    NSEW = tkcontants.NSEW
    CENTER = tkcontants.CENTER

    '''

class tkexpand:
    YES = True
    NO = False

class tkfill:
    from tkinter.constants import (X,Y,BOTH)
    '''
    X = tkcontants.X
    Y = tkcontants.Y
    BOTH = tkcontants.BOTH
    '''

class tkjustify:
    from tkinter.constants import (LEFT,CENTER,RIGHT)


class tkside:
    from tkinter.constants import (LEFT,TOP,RIGHT,BOTTOM)
    '''
    LEFT = tkcontants.LEFT
    TOP = tkcontants.TOP
    RIGHT = tkcontants.RIGHT
    BOTTOM = tkcontants.BOTTOM
    '''

class tksticky:
    from tkinter.constants import (W,E,EW,N,NW,NE,S,SE,SW,NSEW,CENTER,NS)
    '''
    W = tkcontants.W
    E = tkcontants.E
    EW = tkcontants.EW
    N = tkcontants.N
    NW = tkcontants.NW
    NE = tkcontants.NE
    S = tkcontants.S
    SE = tkcontants.SE
    SW = tkcontants.SW
    NSEW = tkcontants.NSEW
    CENTER = tkcontants.CENTER
    '''


class tkorient:
    from tkinter.constants import (HORIZONTAL,VERTICAL)


class tkevent:
    MOUSE_BUTTON_LEFT = '<Button-1>'
    MOUSE_BUTTON_MIDDLE = '<Button-2>'
    MOUSE_BUTTON_RIGHT = '<Button-3>'
    BUTTON_1 = '<Button-1>'
    BUTTON_2 = '<Button-2>'
    class mouse:
        BUTTON_LEFT = '<Button-1>'
        BUTTON_MIDDLE = '<Button-2>'
        BUTTON_RIGHT = '<Button-3>'
    class combobox:
        SELECTED = "<<ComboboxSelected>>"



    


